"""Sub-module with utilities to make experiments easier to write."""

from .compute_node_embedding import (
    compute_node_embedding,
    get_available_node_embedding_methods
)

__all__ = [
    "compute_node_embedding",
    "get_available_node_embedding_methods"
]


"""Abstract class for sequence embedding models."""
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from tensorflow.keras.layers import Embedding  # pylint: disable=import-error
from tensorflow.keras.layers import Input, Layer, GlobalAveragePooling1D
from tensorflow.keras.models import Model  # pylint: disable=import-error
from tensorflow.keras.optimizers import \
    Optimizer  # pylint: disable=import-error
import tensorflow as tf
from .embedder import Embedder
from .layers import NoiseContrastiveEstimation


class Word2Vec(Embedder):
    """Abstract class for sequence embedding models."""

    def __init__(
        self,
        vocabulary_size: int = None,
        embedding_size: int = None,
        embedding: Union[np.ndarray, pd.DataFrame] = None,
        model_name: str = "Word2Vec",
        optimizer: Union[str, Optimizer] = None,
        window_size: int = 16,
        negative_samples: int = 10
    ):
        """Create new sequence Embedder model.

        Parameters
        -------------------------------------------
        vocabulary_size: int = None,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding_size: int = None,
            Dimension of the embedding.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding: Union[np.ndarray, pd.DataFrame] = None,
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        model_name: str = "Word2Vec",
            Name of the model.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        window_size: int = 16,
            Window size for the local context.
            On the borders the window size is trimmed.
        negative_samples: int = 10,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        """
        self._model_name = model_name
        self._window_size = window_size
        self._negative_samples = negative_samples
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            optimizer=optimizer
        )

    def _get_true_input_length(self) -> int:
        """Return length of true input layer."""
        raise NotImplementedError((
            "The method '_get_true_input_length' "
            "must be implemented in child class."
        ))

    def _get_true_output_length(self) -> int:
        """Return length of true output layer."""
        raise NotImplementedError((
            "The method '_get_true_output_length' "
            "must be implemented in child class."
        ))

    def _sort_input_layers(
        self,
        true_input_layer: Layer,
        true_output_layer: Layer
    ) -> Tuple[Layer]:
        """Return input layers for training with the same input sequence.

        Parameters
        ----------------------------
        true_input_layer: Layer,
            The input layer that will contain the true input.
        true_output_layer: Layer
            The input layer that will contain the true output.

        Returns
        ----------------------------
        Return tuple with the tuple of layers.
        """
        raise NotImplementedError((
            "The method '_sort_input_layers' "
            "must be implemented in child class."
        ))

    def _build_model(self):
        """Return Node2Vec model."""
        # Creating the inputs layers
        true_input_layer = Input(
            (self._get_true_input_length(), ),
            dtype=tf.int64,
            name="InputWords"
        )

        # Contextual nodes
        true_output_layer = Input(
            (self._get_true_output_length(), ),
            dtype=tf.int64,
            name="InputContexts"
        )

        # Creating the embedding layer for the contexts
        embedding_layer = Embedding(
            # The plus one is needed for the zero padding!
            input_dim=self._vocabulary_size + 1,
            output_dim=self._embedding_size,
            input_length=self._get_true_input_length(),
            name=Embedder.EMBEDDING_LAYER_NAME,
            # embeddings_regularizer=regularizers.l1_l2(l1=1e-5, l2=1e-4),
            # embeddings_constraint=UnitNorm(),
            mask_zero=True
        )

        embedding = embedding_layer(true_input_layer)

        # Executing average of the embeddings and features (if provided)
        mean_embedding = GlobalAveragePooling1D()(
            embedding,
            mask=embedding_layer.compute_mask(true_input_layer)
        )

        # Adding layer that also executes the loss function
        nce_loss = NoiseContrastiveEstimation(
            # The plus one is needed for the zero padding!
            vocabulary_size=self._vocabulary_size + 1,
            embedding_size=self._embedding_size,
            negative_samples=self._negative_samples,
            positive_samples=self._get_true_output_length(),
        )((mean_embedding, true_output_layer))

        # Creating the actual model
        model = Model(
            inputs=self._sort_input_layers(
                true_input_layer,
                true_output_layer
            ),
            outputs=nce_loss,
            name=self._model_name
        )
        return model

    def _compile_model(self) -> Model:
        """Compile model."""
        # No loss function is needed because it is already executed in
        # the NCE loss layer.
        self._model.compile(
            optimizer=self._optimizer
        )

    @property
    def embedding(self) -> np.ndarray:
        """Return model embeddings.

        Raises
        -------------------
        NotImplementedError,
            If the current embedding model does not have an embedding layer.
        """
        # We need to drop the first column (feature) of the embedding
        # curresponding to the indices 0, as this value is reserved for the
        # masked values. The masked values are the values used to fill
        # the batches of the neigbours of the nodes.
        return Embedder.embedding.fget(self)[1:]  # pylint: disable=no-member

    def fit(
        self,
        *args: List,
        epochs: int = 100,
        early_stopping_monitor: str = "loss",
        early_stopping_min_delta: float = 1,
        early_stopping_patience: int = 2,
        early_stopping_mode: str = "min",
        reduce_lr_monitor: str = "loss",
        reduce_lr_min_delta: float = 2,
        reduce_lr_patience: int = 1,
        reduce_lr_mode: str = "min",
        reduce_lr_factor: float = 0.9,
        verbose: int = 2,
        **kwargs: Dict
    ) -> pd.DataFrame:
        """Return pandas dataframe with training history.

        Parameters
        -----------------------
        *args: List,
            List of parameters to pass to the fit call.
        epochs: int = 100,
            Epochs to train the model for.
        early_stopping_monitor: str = "loss",
            Metric to monitor for early stopping.
        early_stopping_min_delta: float = 0.1,
            Minimum delta of metric to stop the training.
        early_stopping_patience: int = 5,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which trigger early stopping.
        early_stopping_mode: str = "min",
            Direction of the variation of the monitored metric for early stopping.
        reduce_lr_monitor: str = "loss",
            Metric to monitor for reducing learning rate.
        reduce_lr_min_delta: float = 1,
            Minimum delta of metric to reduce learning rate.
        reduce_lr_patience: int = 3,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which reducing learning rate.
        reduce_lr_mode: str = "min",
            Direction of the variation of the monitored metric for learning rate.
        reduce_lr_factor: float = 0.9,
            Factor for reduction of learning rate.
        verbose: int = 2,
            Wethever to show the loading bar.
            Specifically, the options are:
            * 0 or False: No loading bar.
            * 1 or True: Showing only the loading bar for the epochs.
            * 2: Showing loading bar for both epochs and batches.
        **kwargs: Dict,
            Additional kwargs to pass to the Keras fit call.

        Returns
        -----------------------
        Dataframe with training history.
        """
        return super().fit(
            *args,
            epochs=epochs,
            early_stopping_monitor=early_stopping_monitor,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_patience=early_stopping_patience,
            early_stopping_mode=early_stopping_mode,
            reduce_lr_monitor=reduce_lr_monitor,
            reduce_lr_min_delta=reduce_lr_min_delta,
            reduce_lr_patience=reduce_lr_patience,
            reduce_lr_mode=reduce_lr_mode,
            reduce_lr_factor=reduce_lr_factor,
            verbose=verbose,
            **kwargs
        )



"""Sub-module with methods to compute node-embedding with a one-liner."""
from typing import Dict, List, Union, Tuple

import inspect
import pandas as pd
import tensorflow as tf
from cache_decorator import Cache
from ensmallen_graph import EnsmallenGraph


from ..embedders import GraphCBOW, GraphGloVe, GraphSkipGram, Embedder

SUPPORTED_NODE_EMBEDDING_METHODS = {
    "CBOW": GraphCBOW,
    "GloVe": GraphGloVe,
    "SkipGram": GraphSkipGram
}


def get_available_node_embedding_methods() -> List[str]:
    """Return list of supported node embedding methods."""
    global SUPPORTED_NODE_EMBEDDING_METHODS
    return list(SUPPORTED_NODE_EMBEDDING_METHODS.keys())


def get_node_embedding_method(node_embedding_method_name: str) -> Embedder:
    """Return node embedding method curresponding to given name."""
    global SUPPORTED_NODE_EMBEDDING_METHODS
    return SUPPORTED_NODE_EMBEDDING_METHODS[node_embedding_method_name]


def is_node_embedding_method_supported(node_embedding_method_name: str) -> bool:
    """Return boolean value representing if given node embedding method is supported.

    Parameters
    --------------------
    node_embedding_method_name: str,
        Name of the node embedding method.

    Returns
    --------------------
    Whether the given node embedding method is supported.
    """
    return node_embedding_method_name in get_available_node_embedding_methods()


@Cache(
    cache_path=[
        "node_embeddings/{node_embedding_method_name}/{graph_name}/{_hash}_embedding.csv.xz",
        "node_embeddings/{node_embedding_method_name}/{graph_name}/{_hash}_training_history.csv.xz",
    ],
    args_to_ignore=["devices", "verbose"]
)
def _compute_node_embedding(
    graph: EnsmallenGraph,
    graph_name: str,  # pylint: disable=unused-argument
    node_embedding_method_name: str,
    fit_kwargs: Dict,
    verbose: bool = True,
    devices: Union[List[str], str] = None,
    **kwargs: Dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return embedding computed with required node embedding method.

    Specifically, this method also caches the embedding automatically.

    Parameters
    --------------------------
    graph: EnsmallenGraph,
        The graph to embed.
    graph_name: str,
        The name of the graph.
    node_embedding_method_name: str,
        The name of the node embedding method to use.
    fit_kwargs: Dict,
        Arguments to pass to the fit call.
    verbose: bool = True,
        Whether to show loading bars.
    devices: Union[List[str], str] = None,
        The devices to use.
        If None, all GPU devices available are used.
    **kwargs: Dict,
        Arguments to pass to the node embedding method constructor.
        Read the documentation of the selected method.

    Returns
    --------------------------
    Tuple with node embedding and training history.
    """
    # Since the verbose kwarg may be provided also on the fit_kwargs
    # we normalize the parameter to avoid collisions.
    verbose = fit_kwargs.pop("verbose", verbose)
    strategy = tf.distribute.MirroredStrategy(devices=devices)
    with strategy.scope():
        # Creating the node embedding model
        model = get_node_embedding_method(node_embedding_method_name)(
            graph,
            support_mirror_strategy=True,
            **kwargs
        )
        # Fitting the node embedding model
        history = model.fit(
            verbose=verbose,
            **fit_kwargs
        )
        # Extracting computed embedding
        node_embedding = model.get_embedding_dataframe()
    return node_embedding, history


def compute_node_embedding(
    graph: EnsmallenGraph,
    node_embedding_method_name: str,
    devices: Union[List[str], str] = None,
    fit_kwargs: Dict = None,
    verbose: bool = True,
    automatically_drop_unsupported_parameters: bool = False,
    automatically_enable_time_memory_tradeoffs: bool = True,
    **kwargs: Dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return embedding computed using SkipGram on given graph.

    Parameters
    --------------------------
    graph: EnsmallenGraph,
        Graph to embed.
    node_embedding_method_name: str,
        The name of the node embedding method to use.
    devices: Union[List[str], str] = None,
        The devices to use.
        If None, all GPU devices available are used.
    fit_kwargs: Dict = None,
        Arguments to pass to the fit call.
    verbose: bool = True,
        Whether to show loading bars.
    automatically_drop_unsupported_parameters: bool = False,
        If required, we filter out the unsupported parameters.
        This may be useful when running a suite of experiments with a set of
        parameters and you do not want to bother in dropping the parameters
        that are only supported in a subset of methods.
    automatically_enable_time_memory_tradeoffs: bool = True,
        Whether to activate the time memory tradeoffs automatically.
        Often case, this is something you want enabled on your graph object.
        Since, generally, it is a good idea to enable these while
        computing a node embedding we enable these by default.
    **kwargs: Dict,
        Arguments to pass to the node embedding method constructor.
        Read the documentation of the selected method to learn
        which methods are supported by the selected constructor.

    Returns
    --------------------------
    Tuple with node embedding and training history.
    """
    if not is_node_embedding_method_supported(node_embedding_method_name):
        raise ValueError(
            (
                "The given node embedding method `{}` is not supported. "
                "The supported node embedding methods are `{}`."
            ).format(
                node_embedding_method_name,
                get_available_node_embedding_methods()
            )
        )

    # If the fit kwargs are not given we normalize them to an empty dictionary.
    if fit_kwargs is None:
        fit_kwargs = {}

    # If required, we filter out the unsupported parameters.
    # This may be useful when running a suite of experiments with a set of
    # parameters and you do not want to bother in dropping the parameters
    # that are only supported in a subset of methods.
    if automatically_drop_unsupported_parameters and kwargs:
        # Get the list of supported parameters
        supported_parameter = inspect.signature(
            get_node_embedding_method(node_embedding_method_name).__init__
        ).parameters
        # Filter out the unsupported parameters
        kwargs = {
            key: value
            for key, value in kwargs.items()
            if key in supported_parameter
        }

    # If required we enable the time memory tradeoffs.
    if automatically_enable_time_memory_tradeoffs:
        graph.enable(
            vector_destinations=True,
            vector_outbounds=True
        )

    # If devices are given as a single device we adapt this into a list.
    if isinstance(devices, str):
        devices = [devices]

    # Call the wrapper with cache.
    return _compute_node_embedding(
        graph,
        graph_name=graph.get_name(),
        node_embedding_method_name=node_embedding_method_name,
        fit_kwargs=fit_kwargs,
        verbose=verbose,
        devices=devices,
        **kwargs
    )



"""GraphSkipGram model for graph embedding."""
from typing import Union, Dict
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error
from ensmallen_graph import EnsmallenGraph
from .skipgram import SkipGram
from .node2vec import Node2Vec


class GraphSkipGram(Node2Vec):
    """GraphSkipGram model for graph embedding.

    The SkipGram model for graoh embedding receives a central word and tries
    to predict its contexts. The model makes use of an NCE loss layer
    during the training process to generate the negatives.
    """

    def __init__(
        self,
        graph: EnsmallenGraph,
        embedding_size: int = 100,
        optimizer: Union[str, Optimizer] = None,
        negative_samples: int = 10,
        walk_length: int = 128,
        batch_size: int = 256,
        iterations: int = 16,
        window_size: int = 16,
        return_weight: float = 1.0,
        explore_weight: float = 1.0,
        change_node_type_weight: float = 1.0,
        change_edge_type_weight: float = 1.0,
        max_neighbours: int = None,
        elapsed_epochs: int = 0,
        support_mirror_strategy: bool = False,
        random_state: int = 42,
        dense_node_mapping: Dict[int, int] = None,
    ):
        """Create new sequence Embedder model.

        Parameters
        -------------------------------------------
        graph: EnsmallenGraph,
            Graph to be embedded.
        word2vec_model: Word2Vec,
            Word2Vec model to use.
        embedding_size: int = 100,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = None,
            The optimizer to be used during the training of the model.
            By default, if None is provided, Nadam with learning rate
            set at 0.01 is used.
        window_size: int = 16,
            Window size for the local context.
            On the borders the window size is trimmed.
        negative_samples: int = 10,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        walk_length: int = 128,
            Maximal length of the walks.
        batch_size: int = 256,
            Number of nodes to include in a single batch.
        iterations: int = 16,
            Number of iterations of the single walks.
        window_size: int = 16,
            Window size for the local context.
            On the borders the window size is trimmed.
        return_weight: float = 1.0,
            Weight on the probability of returning to the same node the walk just came from
            Having this higher tends the walks to be
            more like a Breadth-First Search.
            Having this very high  (> 2) makes search very local.
            Equal to the inverse of p in the Node2Vec paper.
        explore_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node
            to the one we're coming from in the random walk
            Having this higher tends the walks to be
            more like a Depth-First Search.
            Having this very high makes search more outward.
            Having this very low makes search very local.
            Equal to the inverse of q in the Node2Vec paper.
        change_node_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node of a
            different type than the previous node. This only applies to
            colored graphs, otherwise it has no impact.
        change_edge_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor edge of a
            different type than the previous edge. This only applies to
            multigraphs, otherwise it has no impact.
        max_neighbours: int = None,
            Number of maximum neighbours to consider when using approximated walks.
            By default, None, we execute exact random walks.
            This is mainly useful for graphs containing nodes with extremely high degrees.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        support_mirror_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        random_state: int = 42,
            The random state to reproduce the training sequence.
        dense_node_mapping: Dict[int, int] = None,
            Mapping to use for converting sparse walk space into a dense space.
            This object can be created using the method (available from the
            graph object created using EnsmallenGraph)
            called `get_dense_node_mapping` that returns a mapping from
            the non trap nodes (those from where a walk could start) and
            maps these nodes into a dense range of values.
        """
        super().__init__(
            graph=graph,
            word2vec_model=SkipGram,
            embedding_size=embedding_size,
            optimizer=optimizer,
            negative_samples=negative_samples,
            walk_length=walk_length,
            batch_size=batch_size,
            iterations=iterations,
            window_size=window_size,
            return_weight=return_weight,
            explore_weight=explore_weight,
            change_node_type_weight=change_node_type_weight,
            change_edge_type_weight=change_edge_type_weight,
            max_neighbours=max_neighbours,
            elapsed_epochs=elapsed_epochs,
            support_mirror_strategy=support_mirror_strategy,
            random_state=random_state,
            dense_node_mapping=dense_node_mapping,
        )



"""GloVe model for graph and words embedding."""
from typing import Dict, List, Union

import pandas as pd
import tensorflow as tf
from tensorflow.keras import backend as K  # pylint: disable=import-error
from tensorflow.keras.layers import (Add, Dot,  # pylint: disable=import-error
                                     Embedding, Flatten, Input)
from tensorflow.keras.models import Model  # pylint: disable=import-error
from tensorflow.keras.optimizers import \
    Optimizer  # pylint: disable=import-error

from .embedder import Embedder


class GloVe(Embedder):
    """GloVe model for graph and words embedding.

    The GloVe model for graoh embedding receives two words and is asked to
    predict its cooccurrence probability.
    """

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        optimizer: Union[str, Optimizer] = None,
        alpha: float = 0.75,
    ):
        """Create new GloVe-based Embedder object.

        Parameters
        ----------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
            By default, if None is provided, Nadam with learning rate
            set at 0.01 is used.
        alpha: float = 0.75,
            Alpha to use for the function.
        """
        self._alpha = alpha
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            optimizer=optimizer
        )

    def _glove_loss(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> float:
        """Compute the glove loss function.

        Parameters
        ---------------------------
        y_true: tf.Tensor,
            The true values Tensor for this batch.
        y_pred: tf.Tensor,
            The predicted values Tensor for this batch.

        Returns
        ---------------------------
        Loss function score related to this batch.
        """
        return K.sum(
            K.pow(K.clip(y_true, 0.0, 1.0), self._alpha) *
            K.square(y_pred - K.log(y_true)),
            axis=-1
        )

    def _build_model(self):
        """Create new Glove model."""
        # Creating the input layers
        input_layers = [
            Input((1,)),
            Input((1,))
        ]

        embedding_layers = [
            Embedding(
                self._vocabulary_size,
                self._embedding_size,
                input_length=1,
                weights=None if self._embedding is None else [
                    self._embedding
                ],
                name=Embedder.EMBEDDING_LAYER_NAME
            )(input_layers[0]),
            Embedding(
                self._vocabulary_size,
                self._embedding_size,
                input_length=1,
            )(input_layers[1])
        ]

        # Creating the dot product of the embedding layers
        dot_product_layer = Dot(axes=2)(embedding_layers)

        # Creating the biases layer
        biases = [
            Embedding(
                self._vocabulary_size,
                1,
                input_length=1
            )(input_layer)
            for input_layer in input_layers
        ]

        # Concatenating with an add the three layers
        prediction = Flatten()(Add()([dot_product_layer, *biases]))

        # Creating the model
        glove = Model(
            inputs=input_layers,
            outputs=prediction,
            name="GloVe"
        )
        
        return glove

    def _compile_model(self) -> Model:
        """Compile model."""
        self._model.compile(
            loss=self._glove_loss,
            optimizer=self._optimizer
        )

    def fit(
        self,
        *args: List,
        epochs: int = 1000,
        early_stopping_monitor: str = "loss",
        early_stopping_min_delta: float = 0.001,
        early_stopping_patience: int = 10,
        early_stopping_mode: str = "min",
        reduce_lr_monitor: str = "loss",
        reduce_lr_min_delta: float = 0.01,
        reduce_lr_patience: int = 10,
        reduce_lr_mode: str = "min",
        reduce_lr_factor: float = 0.9,
        verbose: int = 1,
        **kwargs: Dict
    ) -> pd.DataFrame:
        """Return pandas dataframe with training history.

        Parameters
        -----------------------
        epochs: int = 1000,
            Epochs to train the model for.
        early_stopping_monitor: str = "loss",
            Metric to monitor for early stopping.
        early_stopping_min_delta: float = 0.001,
            Minimum delta of metric to stop the training.
        early_stopping_patience: int = 10,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which trigger early stopping.
        early_stopping_mode: str = "min",
            Direction of the variation of the monitored metric for early stopping.
        reduce_lr_monitor: str = "loss",
            Metric to monitor for reducing learning rate.
        reduce_lr_min_delta: float = 0.01,
            Minimum delta of metric to reduce learning rate.
        reduce_lr_patience: int = 10,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which reducing learning rate.
        reduce_lr_mode: str = "min",
            Direction of the variation of the monitored metric for learning rate.
        reduce_lr_factor: float = 0.9,
            Factor for reduction of learning rate.
        verbose: int = 1,
            Wethever to show the loading bar.
            Specifically, the options are:
            * 0 or False: No loading bar.
            * 1 or True: Showing only the loading bar for the epochs.
            * 2: Showing loading bar for both epochs and batches.
        **kwargs: Dict,
            Additional kwargs to pass to the Keras fit call.

        Raises
        -----------------------
        ValueError,
            If given verbose value is not within the available set (-1, 0, 1).

        Returns
        -----------------------
        Dataframe with training history.
        """
        return super().fit(
            *args,
            epochs=epochs,
            early_stopping_monitor=early_stopping_monitor,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_patience=early_stopping_patience,
            early_stopping_mode=early_stopping_mode,
            reduce_lr_monitor=reduce_lr_monitor,
            reduce_lr_min_delta=reduce_lr_min_delta,
            reduce_lr_patience=reduce_lr_patience,
            reduce_lr_mode=reduce_lr_mode,
            reduce_lr_factor=reduce_lr_factor,
            verbose=verbose,
            **kwargs
        )



"""Module with models for graph and text embedding and their Keras Sequences."""
from .embedders import (CBOW, GloVe, GraphCBOW, GraphGloVe, GraphSkipGram,
                        SkipGram)
from .sequences import Word2VecSequence
from .transformers import (CorpusTransformer, EdgeTransformer,
                           GraphTransformer, LinkPredictionTransformer,
                           NodeTransformer)
from .visualizations import GraphVisualization

__all__ = [
    "CBOW",
    "SkipGram",
    "GloVe",
    "GraphCBOW",
    "GraphSkipGram",
    "GraphGloVe",
    "Word2VecSequence",
    "NodeTransformer",
    "EdgeTransformer",
    "GraphTransformer",
    "CorpusTransformer",
    "LinkPredictionTransformer",
    "GraphVisualization"
]



from typing import Dict

import pandas as pd
from ensmallen_graph import EnsmallenGraph
from tensorflow.keras.layers import Dense, Input, Concatenate
from tensorflow.keras.models import Model
from extra_keras_metrics import get_standard_binary_metrics

from ..sequences import EdgePredictionDegreeSequence


class DegreePerceptron:

    def __init__(self, optimizer: str = "nadam"):
        """Create new Perceptron object."""
        self._model_name = self.__class__.__name__
        sources_input = Input((1,), name="SourceNodeDegreeInput")
        destination_input = Input((1,), name="DestinationNodeDegreeInput")
        concatenation = Concatenate()([
            sources_input,
            destination_input
        ])
        output = Dense(
            units=1,
            activation="sigmoid",
        )(concatenation)
        self._model = Model(
            inputs=[
                sources_input,
                destination_input
            ],
            outputs=output,
            name=self._model_name
        )
        self._model.compile(
            loss="binary_crossentropy",
            optimizer=optimizer,
            metrics=get_standard_binary_metrics(),
        )

    def fit(
        self,
        graph: EnsmallenGraph,
        batch_size: int = 2**18,
        batches_per_epoch: int = 2**10,
        negative_samples: float = 1.0,
        ** kwargs: Dict
    ) -> pd.DataFrame:
        """Train model and return training history dataframe.

        Parameters
        -------------------
        graph: EnsmallenGraph,
            Graph object to use for training.
        batch_size: int = 2**18,
            Batch size for the training process.
        batches_per_epoch: int = 2**10,
            Number of batches to train for in each epoch.
        negative_samples: float = 1.0,
            Rate of unbalancing in the batch.
        ** kwargs: Dict,
            Parameteres to pass to the dit call.

        Returns
        --------------------
        Dataframe with traininhg history.
        """
        sequence = EdgePredictionDegreeSequence(
            graph,
            batch_size=batch_size,
            batches_per_epoch=batches_per_epoch,
            negative_samples=negative_samples
        )
        return pd.DataFrame(self._model.fit(
            sequence,
            **kwargs
        ).history)

    def predict(self, *args, **kwargs):
        """Run predict."""
        return self._model.predict(*args, **kwargs)

    def evaluate(self, *args, **kwargs) -> Dict[str, float]:
        """Run predict."""
        return dict(zip(
            self._model.metrics_names,
            self._model.evaluate(*args, **kwargs)
        ))



"""Module with Keras Sequences."""
from .node2vec_sequence import Node2VecSequence
from .edge_prediction_sequence import EdgePredictionSequence
from .edge_prediction_degree_sequence import EdgePredictionDegreeSequence
from .word2vec_sequence import Word2VecSequence
from .node_label_neighbours_sequence import NodeLabelNeighboursSequence

__all__ = [
    "Node2VecSequence",
    "EdgePredictionSequence",
    "EdgePredictionDegreeSequence",
    "Word2VecSequence",
    "NodeLabelNeighboursSequence"
]



"""Keras Sequence for running Neural Network on graph node-label embedding."""
from typing import Tuple

import numpy as np
from ensmallen_graph import EnsmallenGraph
from keras_mixed_sequence import VectorSequence


class NodeLabelNeighboursSequence(VectorSequence):
    """Keras Sequence for running Neural Network on graph node-label prediction."""

    def __init__(
        self,
        graph: EnsmallenGraph,
        max_neighbours: int = None,
        include_central_node: bool = True,
        batch_size: int = 2**8,
        elapsed_epochs: int = 0,
        shuffle: bool = True,
        random_state: int = 42,
        support_mirror_strategy: bool = False
    ):
        """Create new NoLaN sequence.

        Parameters
        --------------------------------
        graph: EnsmallenGraph,
            The graph from which to sample the edges.
        max_neighbours: int = None,
            Number of neighbours to consider.
            If None, the graph mean is used.
        include_central_node: bool = False,
            Whether to include the central node.
            In our experiments, this lead to overfitting.
        batch_size: int = 2**8,
            The batch size to use.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        shuffle: bool = True,
            Whether to shuffle data.
        random_state: int = 42,
            The random state to use to make extraction reproducible.
        support_mirror_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        """
        self._graph = graph
        if max_neighbours is None:
            max_neighbours = self._graph.degrees_median()
        self._max_neighbours = max_neighbours
        self._include_central_node = include_central_node
        self._support_mirror_strategy = support_mirror_strategy
        super().__init__(
            batch_size=batch_size,
            vector=np.array([
                node_id
                for node_id, node_type in enumerate(self._graph.get_node_types())
                if node_type is not None
            ]),
            random_state=random_state,
            shuffle=shuffle,
            elapsed_epochs=elapsed_epochs,
        )

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        neighbours, labels = self._graph.get_node_label_prediction_tuple_by_node_ids(
            node_ids=super().__getitem__(idx),
            random_state=self._random_state + idx,
            include_central_node=self._include_central_node,
            offset=1,
            max_neighbours=self._max_neighbours
        )

        if self._support_mirror_strategy:
            return neighbours.astype(float), labels
        return neighbours, labels
